html,body,#app{margin:0;padding:0;height:100vh;width:100vw;overflow:hidden}#app{display:flex;flex-direction:column}.navbar{padding:10px}.content{width:100%;overflow:hidden;display:flex;flex-grow:1;margin:0;padding:0}.table{margin:0;padding:0}.tabulator-row{padding-left:0!important;padding-right:0!important}.tabulator-cell>a{color:#5f9ea0;text-decoration:none;font-weight:400}.tabulator-cell>a:hover{font-weight:400;text-shadow:1px 0px 0px cadetblue}.targs-cell{width:50px;white-space:nowrap;overflow:hidden;text-overflow:ellipsis}.pagination{margin:1em 0}:root{font-family:Inter,system-ui,Avenir,Helvetica,Arial,sans-serif;line-height:1.5;font-weight:400;color-scheme:light dark;color:#ffffffde;background-color:#242424;font-synthesis:none;text-rendering:optimizeLegibility;-webkit-font-smoothing:antialiased;-moz-osx-font-smoothing:grayscale}a{font-weight:500;color:#0ff;text-decoration:inherit}a:hover{color:#535bf2;cursor:pointer}body{margin:0;display:flex;place-items:center;min-width:320px;min-height:100vh}h1{font-size:3.2em;line-height:1.1}button{border-radius:8px;border:1px solid transparent;padding:.6em 1.2em;font-size:1em;font-weight:500;font-family:inherit;background-color:#1a1a1a;cursor:pointer;transition:border-color .25s}button:hover{border-color:#646cff}button:focus,button:focus-visible{outline:4px auto -webkit-focus-ring-color}@media (prefers-color-scheme: light){:root{color:#213547;background-color:#fff}a:hover{color:#747bff}button{background-color:#f9f9f9}}.tabulator-row:hover,.tabulator-row.tabulator-selectable:hover,.tabulator-row .editable:hover{cursor:auto}.nav-link.active{color:#141414!important;font-size:1em;text-shadow:mediumpurple 1px 1px}
